// This header must be visible to all MySQL code that:
// 1) creates threads, enters mutexes and uses other inter-thread synchronization
// 2) uses GOV_LOG() to write '/var/log/dbgovernor-mysqld.log'.
// Usually it's #include'd into 'my_thread.h'.
// Requires 'global.c.inc' in the project.

// ACHTUNG: Use ONLY 'GOV_...' MACROS in С code patches!

#pragma once

// --------------------------------------------------------------
// Logging to '/var/log/dbgovernor-mysqld.log' from the patch code.
// Part of so-called "extended logging".
//
#include <governor-mysql/log_defs.h>

extern t_write_log_ex *libgovernor_write_log_ex;

// Usage: GOV_LOG(L_ERR, "blabla %d", 235)
#define GOV_LOG(tags, fmt, ...) do { \
		const char *failure = NULL; \
		if (!libgovernor_write_log_ex) /* libgovernor.so not found or old */ \
			failure = "Governor logging unavailable"; \
		else if (libgovernor_write_log_ex(tags, 1, __FILE__, __LINE__, __FUNCTION__, fmt, ##__VA_ARGS__)) /* governor not connected, or "/var/log/dbgovernor-mysqld.log" failed to open, or internal error */ \
			failure = "Governor logging failed"; \
		if (failure && ((tags) & _L_MASK_ALWAYS_ENABLED)) /* failed, but important, let's print by other means */ \
		{ \
			if ((tags) & _L_MASK_ERRORS) \
				sql_print_error(		"(%s) " fmt, failure, ##__VA_ARGS__); \
			else \
				sql_print_information(	"(%s) " fmt, failure, ##__VA_ARGS__); \
		} \
	} while(0)


// --------------------------------------------------------------
// Interception of MySQL synchronization primitives to integrate them with LVE

// LVE slot reservation.
extern void (*my_reserve_slot)();
extern void (*my_release_slot)();
// Use these macros where 'GOV_LVE_MUTEX_...()' and 'GOV_LVE_...THREAD_CREATE()' are inapplicable:
#define GOV_LVE_SLOT_RESERVE()			my_reserve_slot()
#define GOV_LVE_SLOT_RELEASE()			my_release_slot()

// pthread_mutex_...() substitution:
extern int (*my_pthread_lvemutex_lock)(pthread_mutex_t *mutex);
extern int (*my_pthread_lvemutex_trylock)(pthread_mutex_t *mutex);
extern int (*my_pthread_lvemutex_unlock)(pthread_mutex_t *mutex);
#define GOV_LVE_MUTEX_LOCK(mutex)		my_pthread_lvemutex_lock(mutex)
#define GOV_LVE_MUTEX_TRYLOCK(mutex)	my_pthread_lvemutex_trylock(mutex)
#define GOV_LVE_MUTEX_UNLOCK(mutex)		my_pthread_lvemutex_unlock(mutex)

// pthread_create() substitution,
// used in MariaDB and older MySQL:
#define GOV_LVE_PTHREAD_CREATE_DECL() \
	int lve_pthread_create(pthread_t *thread, const pthread_attr_t *attr, void *(*routine)(void*), void *arg)
#define GOV_LVE_PTHREAD_CREATE_IMPL() \
	int lve_pthread_create(pthread_t *__restrict thread, const pthread_attr_t *__restrict attr, void *(*routine)(void*), void *__restrict arg) \
	{ \
		my_reserve_slot(); \
		int rc = pthread_create(thread, attr, routine, arg); \
		my_release_slot(); \
		return rc; \
	}
#define GOV_LVE_PTHREAD_CREATE(thread, attr, routine, arg)		lve_pthread_create(thread, attr, routine, arg)

// my_thread_create() substitution,
// used in modern MySQL:
#define GOV_LVE_MY_THREAD_CREATE_DECL() \
	int lve_my_thread_create(my_thread_handle *thread, const my_thread_attr_t *attr, my_start_routine func, void *arg)
#define GOV_LVE_MY_THREAD_CREATE_IMPL() \
	int lve_my_thread_create(my_thread_handle *thread, const my_thread_attr_t *attr, my_start_routine func, void *arg) \
	{ \
		my_reserve_slot(); \
		int rc = my_thread_create(thread, attr, func, arg); \
		my_release_slot(); \
		return rc; \
	}
#define GOV_LVE_MY_THREAD_CREATE(thread, attr, routine, arg)	lve_my_thread_create(thread, attr, routine, arg)

