
// Implementation for 'main.h'.
// Must be included in every project using that header, usually at the bottom of 'mysqld.cc'.

#if !defined(GOV_CFG_SUPPORT_FULL_LVE) && !defined(GOV_CFG_SUPPORT_LIGHT_LVE)
	#error at least one of GOV_CFG_SUPPORT_FULL_LVE and GOV_CFG_SUPPORT_LIGHT_LVE macros should be defined
#endif


static void *libgovernor_library_handle = NULL;

// used only in this file - for Governor init/deinit
static int (*libgovernor_connect_to_server)() = NULL;
static int (*libgovernor_close_sock)() = NULL;

// used outside of this file
int (*libgovernor_send_info_begin)(const char*) = NULL;
int (*libgovernor_send_info_end)(const char*) = NULL;

static void set_libgovernor_main_funcs_null()
{
	libgovernor_connect_to_server = NULL;
	libgovernor_close_sock = NULL;
	libgovernor_send_info_begin = NULL;
	libgovernor_send_info_end = NULL;
	libgovernor_write_log_ex = NULL;
}

#ifdef __cplusplus
extern "C" {
#endif

static void * (*libgovernor_load_lve_library)() = NULL;			// load library by dl and initialize all pointers to fuctions
static int (*libgovernor_init_lve)() = NULL;					// init_lve wrapper
static void (*libgovernor_destroy_lve)() = NULL;				// destroy_lve wrapper
static int (*libgovernor_enter_lve)(uint32_t *, const char *) = NULL;	// lve_enter wrapper
static int (*libgovernor_enter_lve_light)(uint32_t *) = NULL;	// lve_enter with stored pcookie and UID in thread storage
static void (*libgovernor_lve_exit)(uint32_t *) = NULL;			// lve_leave wrapper

#ifdef __cplusplus
}
#endif

static void set_libgovernor_lve_funcs_null()
{
	libgovernor_load_lve_library = NULL;
	libgovernor_init_lve = NULL;
	libgovernor_destroy_lve = NULL;
	libgovernor_enter_lve = NULL;
	libgovernor_enter_lve_light = NULL;
	libgovernor_lve_exit = NULL;
}


volatile int governor_get_command = 0;

void set_governor_variable()
{
	governor_get_command = 1;
}

void set_governor_variable_reconn()
{
	governor_get_command = 1;
	if (libgovernor_close_sock)
		libgovernor_close_sock();
	if (libgovernor_connect_to_server)
	{
		if (!libgovernor_connect_to_server())
			sql_print_information("Governor reconnected");
		else
			sql_print_error("Governor not reconnected. Failed connection");
	}
}

void set_governor_variable_lve()
{
	governor_get_command = 2;
}

void set_governor_variable_reconn_lve()
{
	set_governor_variable_reconn();
	governor_get_command = 2;
}

#undef pthread_mutex_lock
#undef pthread_mutex_trylock
#undef pthread_mutex_unlock

//Thread dependent variable for thread cookie storage needs for libgovernor_enter_lve, libgovernor_lve_exit
static __thread uint32_t lve_cookie = 0;

int (*put_in_lve)(const char*) = NULL;
void (*lve_thr_exit)() = NULL;


#ifdef GOV_CFG_SUPPORT_LIGHT_LVE	// light implementation of LVE functionality

static int put_in_lve_light(const char *user)
{
	return libgovernor_enter_lve(&lve_cookie, user);
}

static void lve_thr_exit_light()
{
	libgovernor_lve_exit(&lve_cookie);
}

static int my_pthread_lvemutex_lock_light(pthread_mutex_t *mp)
{
	return pthread_mutex_lock(mp);
}

static int my_pthread_lvemutex_trylock_light(pthread_mutex_t *mutex)
{
	return pthread_mutex_trylock(mutex);
}

static int my_pthread_lvemutex_unlock_light(pthread_mutex_t *mutex)
{
	return pthread_mutex_unlock(mutex);
}

static void my_reserve_slot_light()
{
}

static void my_release_slot_light()
{
}

static void lve_setup_light_functionality()
{
	put_in_lve = put_in_lve_light;
	lve_thr_exit = lve_thr_exit_light;
	my_reserve_slot = my_reserve_slot_light;
	my_release_slot = my_release_slot_light;
	my_pthread_lvemutex_lock = my_pthread_lvemutex_lock_light;
	my_pthread_lvemutex_trylock = my_pthread_lvemutex_trylock_light;
	my_pthread_lvemutex_unlock = my_pthread_lvemutex_unlock_light;
}

#endif // GOV_CFG_SUPPORT_LIGHT_LVE


#ifdef GOV_CFG_SUPPORT_FULL_LVE	// full implementation of LVE functionality

// Thread state - the main thing distinguishing Full mode from the Light one.
// Let's use as short names as possible - it's safe enough when we're #include'd at the bottom.
enum
{
	OFF					= 0,	// default thread state, before entering LVE
	ON					= 1		// after entering LVE

// values > 2 mean "off in nested mutexes":
//	ON+1				= 2		// after locking 1st mutex (outer) and temporarily exiting LVE
//	ON+2				= 3		// after locking 2nd mutex (nested)
//	ON+3				= 4		// after locking 3rd mutex (nested)
//						...
};
static __thread int lve = OFF;

// Transition scheme:
// [thread start]  ==>  OFF  ==>  [enter LVE]  ==>  ON  ==>  [exit  LVE temporarily]==>[  lock 1st mutex]  ==>  ON+1  ==>  [  lock 2nd mutex]  ==>  ON+2  ==>  ...
// [thread end]    <==  OFF  <==  [exit  LVE]  <==  ON  <==  [enter LVE back]       <==[unlock 1st mutex]  <==  ON+1  <==  [unlock 2nd mutex]  <==  ON+2  <==  ...
//               ^ ^ ^ ^ ^ ^                       ^^^        ^ ^ ^ ^ ^...
//              (outside LVE)                    (in LVE)    (mutex pit > > >)

// We should be in LVE if and only if lve==ON. If it's >ON, it's essentially !=ON. It's "mutex pit".

int put_in_lve_full(const char *user)
{
	if (lve != OFF)
		GOV_LOG(L_ERRSENTRY|L_LVE, "thread looks like already in LVE");
	lve = OFF;
	int ret = libgovernor_enter_lve(&lve_cookie, user);
	if (ret)
		return ret;
	lve = ON;
	return 0;
}

void lve_thr_exit_full()
{
	if (lve != ON)
		GOV_LOG(L_ERR|L_LVE, "thread looks like already outside of LVE");
	if (lve == ON)
		libgovernor_lve_exit(&lve_cookie);		// ONLY if things are clear
	lve = OFF;		// in ANY case
}

// The below 3 functions are our substitutes for 'pthread_mutex_...()' API family.
// Inside them, apart from calling the original API, we maintain 'lve' value.
// We do nothing when we're outside LVE (lve==OFF). So, if we're OFF initially, these functions won't ever change it.
// But if we became ON, we start to increment 'lve' on each mutex lock and decrement on each unlock.
// This way, when we transit from ON to ON+1, we know that this increment is caused by the outer (or 'first in the stack') mutex.
// At this moment, we leave LVE for the whole duration of outer mutex locked state - to avoid being LVE-frozen while holding global resources.
// Increment from ON+1 to ON+2, and so on, mean locking nested mutexes (i.e., locking more before unlocking the initial one). In these cases, we do nothing.
// Decrement from ON+2 to ON+1 also means unlocking nested ones.
// But when we decrement from ON+1 to ON, this means that this unlock is done on the outer mutex. And we re-enter LVE.
//
// Our solution works only under Assumption #1 - that the code covered with LVE always locks and releases mutexes in LIFO order.
// If the LVE-protected 3'rd party code:
// 1) is a well-defined piece of business logic
// 2) free of bugs
// 3) uses only trivial IPC schemes,
// then most probably our assumption holds.
// For now, we LVE-wrap only a single MySQL/MariaDB call, and it looks OK.
//
// But the efficiency of our solution depends also on Assumption #2 -
// that the time spent inside those mutexes is significant.
// If the mutex protects a very fast thing, like a reference counter increment, our overhead can be intolerable.

int my_pthread_lvemutex_lock_full(pthread_mutex_t *mutex)
{
	if (lve == ON)							// in LVE
		libgovernor_lve_exit(&lve_cookie);		// temporarily exit, at least for the duration of locking
	int ret = pthread_mutex_lock(mutex);
	if (!ret)
	{
		if (lve >= ON)						// in LVE or in mutex pit
			lve++;							// step deeper into mutex pit
	} else
		if (lve == ON && libgovernor_enter_lve_light(&lve_cookie))		// lock failed, need to re-enter, failed to do so
		{
			lve = OFF;		// prematurely finish current LVE invocation for the thread
			GOV_LOG(L_ERR|L_LVE, "failed to re-enter after unsuccessful mutex lock attempt");
		}
	return ret;
}

int my_pthread_lvemutex_unlock_full(pthread_mutex_t *mutex)
{
	int ret = pthread_mutex_unlock(mutex);
	if (!ret && lve > ON)	// successfully unlocked a mutex, while being in mutex pit
	{
		lve--;								// step up from mutex pit,
		if (lve == ON && libgovernor_enter_lve_light(&lve_cookie))		// need to re-enter, failed to do so
		{
			lve = OFF;		// prematurely finish current LVE invocation for the thread
			GOV_LOG(L_ERR|L_LVE, "failed to re-enter after mutex unlock");
		}
	}
	return ret;
}

int my_pthread_lvemutex_trylock_full(pthread_mutex_t *mutex)
{
	if (lve == ON)								// in LVE
		libgovernor_lve_exit(&lve_cookie);			// temporarily exit, at least for the duration of locking
	int ret = pthread_mutex_trylock(mutex);
	if (!ret)
	{
		if (lve >= ON)							// in LVE or in mutex pit
			lve++;								// step deeper into mutex pit
	} else
		if (lve == ON && libgovernor_enter_lve_light(&lve_cookie))		// lock failed, need to re-enter, failed to do so
		{
			lve = OFF;		// prematurely finish current LVE invocation for the thread
			GOV_LOG(L_ERR|L_LVE, "failed to re-enter after unsuccessful mutex trylock");
		}
	return ret;
}

// Same as successful mutex lock.
void my_reserve_slot_full()
{
	if (lve == ON)
		libgovernor_lve_exit(&lve_cookie);
	if (lve >= ON)
		lve++;
}

// Same as successful mutex unlock.
void my_release_slot_full()
{
	if (lve > ON)
	{
		lve--;
		if (lve == ON && libgovernor_enter_lve_light(&lve_cookie))
		{
			lve = OFF;			// prematurely finish current LVE invocation for the thread
			GOV_LOG(L_ERR|L_LVE, "failed to re-enter after slot release");
		}
	}
}

static void lve_setup_full_functionality()
{
	put_in_lve = put_in_lve_full;
	lve_thr_exit = lve_thr_exit_full;
	my_reserve_slot = my_reserve_slot_full;
	my_release_slot = my_release_slot_full;
	my_pthread_lvemutex_lock = my_pthread_lvemutex_lock_full;
	my_pthread_lvemutex_trylock = my_pthread_lvemutex_trylock_full;
	my_pthread_lvemutex_unlock = my_pthread_lvemutex_unlock_full;
}

#endif // GOV_CFG_LVE_FULL_MODE

void governor_init()
{
	governor_get_command = 0;

	set_libgovernor_main_funcs_null();
	set_libgovernor_lve_funcs_null();

	const char *path = "libgovernor.so";
	libgovernor_library_handle = dlopen(path, RTLD_LAZY);
	sql_print_information("%s %s", path, libgovernor_library_handle ? "found" : "not found");

	#define LOAD_LIBGOVERNOR_FUNC(fptr, ftype, fname)\
		if (!load_error)\
		{\
			fptr = (ftype)dlsym(libgovernor_library_handle, fname);\
			load_error = dlerror();\
			if (load_error)\
				sql_print_warning("Failed to find Governor function '%s': %s", fname, load_error);\
		}

	if (libgovernor_library_handle)
	{
		const char *load_error = 0;

		// It's IMPORTANT that function pointer variables here have DIFFERENT names from the corresponding functions in libgovernor.so.
		// Because dlopen(..., RTLD_LAZY) hides pitfalls. Imagine:
		// 1) 'send_info_begin()' is defined in 'libgovernor.so', and is called somewhere inside it;
		// 2) 'mysqld' executable has a variable 'send_info_begin';
		// 3) 'mysqld' calls dlopen("libgovernor.so", RTLD_LAZY), and everything is alright UNTIL
		// 4) a call to 'send_info_begin()' is done from inside of 'libgovernor.so',
		//    postponed symbol resolution resolves 'send_info_begin()' function address in 'libgovernor.so'
		//    to the address of 'send_info_begin' variable in 'mysqld', and the call crashes.
		//
		// I believe the problem exists with ANY external symbol clash between the exe and .so,
		// but it becomes EXTREMELY dangerous with RTLD_LAZY.

		LOAD_LIBGOVERNOR_FUNC(libgovernor_connect_to_server,	int (*)(),				"connect_to_server");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_send_info_begin,		int (*)(const char*),	"send_info_begin");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_send_info_end,		int (*)(const char*),	"send_info_end");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_close_sock,			int (*)(),				"close_sock");

		if (load_error)
		{
			set_libgovernor_main_funcs_null();
			sql_print_warning("Governor functionality disabled");
		} else
		{
			sql_print_information("All main Governor functions found");

			LOAD_LIBGOVERNOR_FUNC(libgovernor_write_log_ex, t_write_log_ex*, "write_log_ex");
			if (load_error)
				libgovernor_write_log_ex = 0;
			sql_print_information("Governor logging %s", libgovernor_write_log_ex ? "available" : "unavailable");
		}
	}

	if (libgovernor_connect_to_server)
	{
		if (!libgovernor_connect_to_server())
			sql_print_information("Governor connected");
		else
			sql_print_error("Governor not connected");
	}

	if (libgovernor_library_handle)
	{
		const char *load_error = 0;

		LOAD_LIBGOVERNOR_FUNC(libgovernor_load_lve_library,	void * (*)(),						"governor_load_lve_library");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_init_lve,			int (*)(),							"governor_init_lve");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_destroy_lve,		void (*)(),							"governor_destroy_lve");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_enter_lve,		int (*)(uint32_t *, const char *),	"governor_enter_lve");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_enter_lve_light,	int (*)(uint32_t *),				"governor_enter_lve_light");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_lve_exit,			void (*)(uint32_t *),				"governor_lve_exit");

		if (load_error)
		{
			set_libgovernor_lve_funcs_null();
			sql_print_warning("Governor LVE functionality disabled");
		} else
		{
			sql_print_information("All Governor LVE functions found");
		}
	}

	#undef LOAD_LIBGOVERNOR_FUNC

	if (libgovernor_load_lve_library)
	{
		if (!libgovernor_load_lve_library())	// this one returns 0 on _failure_
		{
			sql_print_information("Can't get LVE functions");
			set_libgovernor_lve_funcs_null();
		} else if (libgovernor_init_lve())
		{
			sql_print_error("Governor LVE initialization error");
			set_libgovernor_lve_funcs_null();
		}
	}

	const char *lveMode = "";
	if (!libgovernor_load_lve_library)
		lveMode = "no";
	else
	{
#if defined(GOV_CFG_SUPPORT_LIGHT_LVE) && defined(GOV_CFG_SUPPORT_FULL_LVE)
		const char *lightFile = "/usr/share/lve/dbgovernor/cl_light_lve";
		struct stat st;
		bool light = stat(lightFile, &st) >= 0;
		sql_print_information("'%s' %s", lightFile, light ? "found" : "not found");
		if (light)
			lve_setup_light_functionality();
		else
			lve_setup_full_functionality();
		lveMode = light ? "light" : "full";
#elif defined(GOV_CFG_SUPPORT_LIGHT_LVE)
		lve_setup_light_functionality();
		lveMode = "light (the only supported)";
#elif defined(GOV_CFG_SUPPORT_FULL_LVE)
		lve_setup_full_functionality();
		lveMode = "full (the only supported)";
#else
		#error cannot get here, ensured earlier
#endif
	}
	sql_print_information("Using %s LVE functionality", lveMode);
	GOV_LOG(L_LIFE, "LVE mode: %s", lveMode);
}

void governor_deinit()
{
	if (libgovernor_destroy_lve)
		libgovernor_destroy_lve();

	if (libgovernor_close_sock)
		libgovernor_close_sock();

	if (libgovernor_library_handle)
	{
		dlclose(libgovernor_library_handle);	// calls libgovernor.so->fini_libgovernor()->close_log(), so after this GOV_LOG() is not fully functional
		libgovernor_library_handle = NULL;	// JIC
	}

	// No libgovernor funcs should be called since now, and we are confident enough - we're right before mysqld process exit.
	// But JIC it's always safer to crash on NULL pointers than on garbage:
	set_libgovernor_main_funcs_null();
	set_libgovernor_lve_funcs_null();
}

// Prevent 'pthread_mutex_...' reuse, becuase we #undef'ed it earlier. Anyways we must include this whole file at the bottom, but JIC:
#define pthread_mutex_lock(m)		error: pthread_mutex_...() spoiled; include this file only at the bottom
#define pthread_mutex_trylock(m)	error: pthread_mutex_...() spoiled; include this file only at the bottom
#define pthread_mutex_unlock(m)		error: pthread_mutex_...() spoiled; include this file only at the bottom

