
// This header must be visible to mysqld code that:
// 1) performs startup and shutdown
// 2) enters/exits LVE - e.g., when executing SQL queries.
// Usually it's #include'd into 'mysqld.h'.
// Requires 'main.c.inc' in the project.

// ACHTUNG: Use ONLY 'GOV_...' MACROS in С code patches!

#pragma once

// --------------------------------------------------------------
// Governor initialization and de-initialization:
// loading functions from libgovernor.so, connecting to db_governor, selecting LVE full/light mode, etc.
//
void governor_init();
void governor_deinit();

#define GOV_INIT()		governor_init()
#define GOV_DEINIT()	governor_deinit()


// --------------------------------------------------------------
// libgovernor mode
//
extern volatile int governor_get_command;

// Set libgovernor mode
void set_governor_variable();
void set_governor_variable_reconn();
void set_governor_variable_lve();
void set_governor_variable_reconn_lve();

// Expose libgovernor mode as MySQL variable - to be used in 'SHOW VARIABLES', 'SET', etc.
#define GOV_SQLVARADDR_ENABLE()		(const_cast<int*>(&governor_get_command))	// cast away 'volatile'

// Implement custom MySQL commands, changing libgovernor mode and forcing reconnection to db_governor
#define GOV_SQLCMDIMPL_ENABLE()						set_governor_variable()
#define GOV_SQLCMDIMPL_ENABLE_AND_RECONNECT()		set_governor_variable_reconn()
#define GOV_SQLCMDIMPL_ENABLE_LVE()					set_governor_variable_lve()
#define GOV_SQLCMDIMPL_ENABLE_LVE_AND_RECONNECT()	set_governor_variable_reconn_lve()


// --------------------------------------------------------------
// LVE entering and exit.
//
extern int (*libgovernor_send_info_begin)(const char*);
extern int (*libgovernor_send_info_end)(const char*);

extern int (*put_in_lve)(const char *user);
extern void (*lve_thr_exit)();

#define _GOV_PREPARE_FOR_LVE(thd, mSecCtx, mUserName) \
	int govCmd = governor_get_command;  /* use the SAME snapshot of this volatile for ALL our checks */ \
	/* Make fast and cheap checks first: */ \
	bool govSendInfo = libgovernor_send_info_begin && libgovernor_send_info_end && govCmd;  /* both functions loaded from libgovernor, and Governor functionality enabled */ \
	bool govEnterLVE = put_in_lve && lve_thr_exit && govCmd==2;  /* Governor LVE functionality is properly initialized (mode is 'Light or 'Full', but not 'No'), and it is currently enabled by 'governor_get_command' value */ \
	std::string govUserName;  /* to use in ALL our actions */ \
	/* If anything is required, make more checks: */ \
	if (govSendInfo || govEnterLVE) \
	{ \
		if ((  /* none of Governor custom SQL commands: */ \
			   thd->lex->sql_command != SQLCOM_ENABLE_GOVERNOR \
			&& thd->lex->sql_command != SQLCOM_ENABLE_RECONN_GOVERNOR \
			&& thd->lex->sql_command != SQLCOM_ENABLE_GOVERNOR_LVE \
			&& thd->lex->sql_command != SQLCOM_ENABLE_RECONN_GOVERNOR_LVE) \
			&& /* and non-empty user name: */ \
			(  thd->mSecCtx \
			&& thd->mSecCtx->mUserName \
			&& thd->mSecCtx->mUserName[0])) \
			govUserName = thd->mSecCtx->mUserName;  /* all is OK, save username and do all planned actions on it */ \
		else \
			govSendInfo = govEnterLVE = false;  /* wrong SQL command or empty username, do nothing */ \
	}

#define GOV_SEND_INFO_BEGIN(thd, mSecCtx, mUserName) \
	_GOV_PREPARE_FOR_LVE(thd, mSecCtx, mUserName); \
	if (govSendInfo)\
		libgovernor_send_info_begin(govUserName.c_str());

#define GOV_SEND_INFO_END() \
	if (govSendInfo) \
		libgovernor_send_info_end(govUserName.c_str());

// These two macros are the ultimate purpose of the patch:
#define GOV_LVE_ENTER() \
	int lveEnterResult = -1; \
	if (govEnterLVE) \
	{ \
		lveEnterResult = put_in_lve(govUserName.c_str()); \
		if (lveEnterResult < 0) \
			my_error(ER_GET_ERRNO, MYF(0), errno, "Can't enter into LVE"); \
	}

#define GOV_LVE_EXIT() \
	if (lveEnterResult == 0) \
		lve_thr_exit();
